// Copyright 2011-2013 Kwant authors.
//
// This file is part of Kwant.  It is subject to the license terms in the file
// LICENSE.rst found in the top-level directory of this distribution and at
// http://kwant-project.org/license.  A list of Kwant authors can be found in
// the file AUTHORS.rst at the top-level directory of this distribution and at
// http://kwant-project.org/authors.

#ifndef DEFS_H
#define DEFS_H

#include <stdint.h>

// The integer type which will be used to save node and edge id's.  Must be
// signed.
typedef int32_t gint;

#endif // DEFS_H
